"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var pathModule = require("path");
var tsParser_1 = require("./tsParser");
var ts = require("typescript");
var glob = require("glob");
var Leaf = require("leaf-ts");
var tsFormatter_1 = require("./tsFormatter");
var TSRouteParser = /** @class */ (function () {
    function TSRouteParser(seeds) {
        var todos = [];
        for (var _i = 0, seeds_1 = seeds; _i < seeds_1.length; _i++) {
            var seed = seeds_1[_i];
            if (pathModule.extname(seed) == ".ts") {
                todos.push(seed);
                continue;
            }
            var files = glob.sync(seed + "/**/*.ts");
            todos.push.apply(todos, files);
        }
        this.project = new tsParser_1.TSProject(todos);
        this.seeds = todos;
    }
    // TODO filter out those APISpec not included in the root seeds.
    TSRouteParser.prototype.parse = function (seeds) {
        if (seeds === void 0) { seeds = this.seeds; }
        var reg = /^.*APIService$/;
        var infos = [];
        for (var _i = 0, _a = this.seeds; _i < _a.length; _i++) {
            var seed = _a[_i];
            var file = this.project.getFile(seed);
            if (!file) {
                throw new Error("seed " + seed + " not found");
            }
            var clas = file.getClasses();
            for (var _b = 0, clas_1 = clas; _b < clas_1.length; _b++) {
                var cla = clas_1[_b];
                if (reg.test(cla.name)) {
                    var info = this.getRouteInfoRecursive(file, cla.name);
                    info.seed = seed;
                    infos.push(info);
                }
            }
        }
        return infos;
    };
    TSRouteParser.prototype.extractAPIImplementation = function (file, name) {
        var cla = file.getClass(name);
        var result = {
            class: null,
            initializer: null,
            implementations: []
        };
        if (!cla) {
            return result;
        }
        result.class = cla;
        if (!cla.methods.initialize)
            return result;
        result.initializer = cla.methods.initialize;
        var init = cla.methods.initialize.node;
        var node = file.querySelect(init, "Block");
        var impls = [];
        ts.forEachChild(node, function (child) {
            if (child.kind !== tsParser_1.Kinds.ExpressionStatement)
                return;
            var call = file.querySelect(child, "CallExpression");
            // not a invoke
            if (!call)
                return;
            var spec = file.parseInvoke(call);
            if (!spec.invoke.access || !spec.invoke.access)
                return;
            var accs = spec.invoke.access.accesses;
            var last = accs[accs.length - 1];
            if (last != "api")
                return;
            var name = spec.params[0];
            if (name.literalType != "string") {
                console.error("Invoke api without providing name " + file.source.fileName + ": " + name.node.pos + " ");
                return;
            }
            impls.push({
                spec: spec,
                className: cla.name,
                api: name.literalValue,
                start: child.getFullStart(),
                width: child.getFullWidth(),
                parent: cla.methods.initialize
            });
            //console.log(spec.invoke, spec.params.length)
        });
        result.implementations = impls;
        return result;
    };
    TSRouteParser.prototype.getRouteInfoRecursive = function (file, name) {
        var _this = this;
        var _a, _b;
        // 1. look up class with specific API names recursively look up spec in it's base class
        // 2. look up specs incurrent file and over write
        var info = {
            className: null,
            name: null,
            routes: {},
            prefix: "",
            class: null,
            initializer: null,
            implementations: [],
            parent: null,
            seed: file.source.fileName
        };
        var impl = this.extractAPIImplementation(file, name);
        var cla = impl.class;
        if (!cla)
            return null;
        info.class = cla;
        if (cla.base && cla.base.class) {
            var file_1 = new tsParser_1.TSFile(this.project, cla.base.class.node.getSourceFile());
            var base = this.getRouteInfoRecursive(file_1, cla.base.name);
            if (base) {
                for (var name_1 in base.routes) {
                    info.routes[name_1] = base.routes[name_1];
                }
                info.name = base.name || info.name;
                info.prefix = base.prefix;
                info.parent = base;
                (_a = info.implementations).push.apply(_a, base.implementations);
            }
        }
        (_b = info.implementations).push.apply(_b, impl.implementations);
        if (impl.initializer) {
            info.initializer = impl.initializer;
        }
        info.className = cla.name;
        var specName = info.className + "Spec";
        var nss = file.getNamespaces();
        var ns = nss.find(function (ns) { return ns.name == specName; });
        if (!ns)
            return info;
        var nameVar = ns.block.variables.name;
        info.name = nameVar && nameVar.value && nameVar.value.literalValue || info.name;
        var prefixVar = ns.block.variables.prefix;
        if (prefixVar && prefixVar.value && prefixVar.value.literalType == "string") {
            info.prefix = prefixVar.value.literalValue;
        }
        if (!info.name) {
        }
        var canonicalName = Leaf.Util.slugToCamel(info.name.replace("api.", "").replace(/\./, "-"));
        var apiVar = ns.block.variables.API;
        ts.forEachChild(apiVar.value.node, function (child) {
            var id = file.querySelect(child, "Identifier".split("."));
            var apiSpecClaNode = file.querySelect(child, "NewExpression.ClassExpression".split("."));
            if (!apiSpecClaNode)
                return;
            var apiSpecCla = file.parseClass(apiSpecClaNode);
            var methodProp = apiSpecCla.properties.method;
            var pathProp = apiSpecCla.properties.path;
            var returnProp = apiSpecCla.properties.return;
            var bodyProp = apiSpecCla.properties.body;
            var queryProp = apiSpecCla.properties.query;
            var paramsProp = apiSpecCla.properties.params;
            if (returnProp && !returnProp.type) {
                file.inspectKind(returnProp.node);
                //console.error(returnProp.type.node.getText())
            }
            info.routes[id.text] = {
                name: id.text,
                path: pathProp && pathProp.value && pathProp.value.literalValue || canonicalName + "/" + id.text,
                method: methodProp && methodProp.value && methodProp.value.literalValue || methodProp && methodProp.type && JSON.parse(methodProp.type.literal) || "POST",
                returnTypeString: _this.typeFromProp(returnProp) || null,
                bodyTypeString: _this.typeFromProp(bodyProp) || null,
                queryTypeString: _this.typeFromProp(queryProp) || null,
                paramsTypeString: _this.typeFromProp(paramsProp) || null,
            };
        });
        return info;
    };
    TSRouteParser.prototype.typeFromProp = function (prop) {
        if (!prop)
            return null;
        if (prop.type) {
            return prop.type.literal || prop.type.node.getText();
        }
        if (prop.value) {
            return prop.value.literal;
        }
    };
    return TSRouteParser;
}());
exports.TSRouteParser = TSRouteParser;
var TSRouteFile = /** @class */ (function () {
    function TSRouteFile(file) {
        this.file = file;
    }
    TSRouteFile.prototype.isRouteFile = function () {
    };
    return TSRouteFile;
}());
exports.TSRouteFile = TSRouteFile;
var RouteAPIGenerator = /** @class */ (function () {
    function RouteAPIGenerator(services) {
        this.services = services;
    }
    RouteAPIGenerator.prototype.build = function () {
        var _this = this;
        var header = "import * as Leaf from \"leaf-ts\"\n            ";
        var codes = this.services.map(function (info) { return _this.buildAPIService(info); });
        codes.push(this.buildAPIBundle(this.services));
        var code = codes.join("\n");
        return tsFormatter_1.format(header + code);
    };
    RouteAPIGenerator.prototype.buildAPIBundle = function (services) {
        var apis = services.map(function (info) {
            if (!info.name) {
                console.error("Invalid API INFO", info);
                throw new Error("Invalid API Info");
            }
            var name = Leaf.Util.slugToCamel(info.name.replace("api.", "").replace(/\./g, "-"));
            return "readonly " + name + " = new (Generate" + info.className + " ()) ()";
        }).join("\n");
        return "\nexport class APIService extends Leaf.Service {\n    readonly name = \"APIService\"\n    " + apis + "\n}\nexport function GenerateAPI() {\n    return new APIService()\n}\n";
    };
    RouteAPIGenerator.prototype.buildAPIService = function (info) {
        var apis = Object.keys(info.routes).map(function (key) {
            var api = info.routes[key];
            var requestTypes = [];
            if (api.paramsTypeString) {
                requestTypes.push(api.paramsTypeString);
            }
            if (api.queryTypeString) {
                requestTypes.push(api.queryTypeString);
            }
            if (api.bodyTypeString) {
                requestTypes.push(api.bodyTypeString);
            }
            requestTypes = requestTypes.filter(function (item) { return !!item; });
            return "\n" + key + " = factory.createAPIFunction < " + (requestTypes.join("&") || "{}") + " , " + (api.returnTypeString || "null") + "> ({\n    method: \"" + api.method + "\",\n    path: \"" + api.path + "\"\n})\n    ";
        });
        var result = "\nexport function Generate" + info.className + " <T extends (new (...args: any[]) => {}) = new (...args: any[]) => {}>(factory: Leaf.APIFactory = new Leaf.APIFactory({ prefix: \"" + info.prefix + "\", bodyType: \"json\", serializer: Leaf.Serializer.getDefault() }), Base: T = class { } as any){\n    return class extends Base {\n        public factory = factory\n        " + apis.join("\n") + "\n}\n}";
        return result;
    };
    return RouteAPIGenerator;
}());
exports.RouteAPIGenerator = RouteAPIGenerator;
var RouteAPISyncer = /** @class */ (function () {
    function RouteAPISyncer(parser, info) {
        this.parser = parser;
        this.info = info;
    }
    RouteAPISyncer.prototype.merge = function (option) {
        var _this = this;
        var file = this.parser.project.getFile(option.path);
        if (!file) {
            throw new Error("File not exists");
        }
        var origContent = file.source.getFullText();
        var replacer = new Replacer(origContent);
        if (!this.info.class) {
            return tsFormatter_1.format(replacer.add(origContent.length, this.buildClass()).apply());
        }
        if (!this.info.initializer) {
            var last = this.info.class.children[this.info.class.children.length - 1];
            var offset = last.getFullText().lastIndexOf("}") - 1;
            var start_1 = last.getFullStart() + offset;
            return tsFormatter_1.format(replacer.add(start_1, "\n" + this.buildInitializer()).apply());
        }
        var impls = this.info.implementations;
        var diff = Leaf.Util.arrayDiffTyped(this.info.implementations, Object.keys(this.info.routes), function (a, b) {
            if (a.api == b)
                return true;
            return false;
        });
        if (diff.left.length == 0 && diff.right.length == 0)
            return tsFormatter_1.format(file.source.getFullText());
        if (impls.length == 0) {
            var start_2 = this.info.initializer.block.node.getFullStart();
            var at = this.info.initializer.block.node.getFullText().lastIndexOf("}") - 1;
            replacer.add(start_2 + at, this.buildDefaultImplementations(Object.keys(this.info.routes).map(function (key) { return _this.info.routes[key]; })));
            return tsFormatter_1.format(replacer.apply());
        }
        var block = this.info.initializer.block.node;
        var start = block.getFullText().lastIndexOf("}") + block.getFullStart();
        if (diff.left.length == diff.right.length && diff.right && option.guess) {
            for (var _i = 0, _a = diff.left; _i < _a.length; _i++) {
                var item = _a[_i];
                replacer.delete(item.start, item.width);
            }
        }
        var content = this.buildDefaultImplementations(diff.right.map(function (item) { return _this.info.routes[item]; }));
        replacer.add(start, content);
        return tsFormatter_1.format(replacer.apply());
    };
    RouteAPISyncer.prototype.buildDefaultImplementations = function (infos) {
        return infos.map(function (info) { return "\nthis.api(\"" + info.name + "\", ctx => {\n    ctx.error(new Error(\"API " + info.name + " not implemented\"))\n})\n    "; }).join("\n");
    };
    RouteAPISyncer.prototype.buildFile = function (routePath, path) {
        var routeSpecPath = pathModule.relative(pathModule.dirname(path), routePath).replace(".ts", "");
        return tsFormatter_1.format("import * as RouteSpec from \"" + routeSpecPath + "\"\nimport * as Leaf from \"leaf-ts\"\n" + this.buildClass() + "\n");
    };
    RouteAPISyncer.prototype.buildClass = function () {
        return "\nexport class $ {this.info.className} extends RouteSpec." + this.info.className + " {\n    " + this.buildInitializer() + "\n}\n";
    };
    RouteAPISyncer.prototype.buildInitializer = function (apis) {
        var _this = this;
        apis = apis || Object.keys(this.info.routes).map(function (key) {
            if (_this.info.implementations.find(function (impl) { return impl.api == key; }))
                return null;
            return _this.info.routes[key];
        }).filter(function (item) { return !!item; });
        var codes = this.buildDefaultImplementations(apis);
        return "initialize(){\n    super.initialize()\n    " + codes + "\n} ";
    };
    return RouteAPISyncer;
}());
exports.RouteAPISyncer = RouteAPISyncer;
var Replacer = /** @class */ (function () {
    function Replacer(content) {
        this.content = content;
        this.todos = [];
    }
    Replacer.prototype.add = function (at, content) {
        this.todos.push({
            start: at,
            width: 0,
            replacement: content
        });
        return this;
    };
    Replacer.prototype.delete = function (at, width) {
        this.todos.push({
            start: at,
            width: width,
            replacement: ""
        });
        return this;
    };
    Replacer.prototype.apply = function () {
        var content = this.content;
        var prev;
        this.todos.sort(function (a, b) { return a.start - b.start; });
        for (var _i = 0, _a = this.todos; _i < _a.length; _i++) {
            var todo = _a[_i];
            if (!prev) {
                prev = todo;
                continue;
            }
            if (todo.start < prev.start + prev.width)
                throw new Error("Invalid replacer build");
        }
        var adjustment = 0;
        for (var _b = 0, _c = this.todos; _b < _c.length; _b++) {
            var todo = _c[_b];
            var start = todo.start + adjustment;
            var end = start + todo.width;
            content = content.slice(0, start) + todo.replacement + content.slice(end);
            adjustment += todo.replacement.length - todo.width;
        }
        return content;
    };
    return Replacer;
}());
//# sourceMappingURL=tsRouteParser.js.map